"""This module contains classes representing the API classes generated by the SDK generator."""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, cast

from cognite.client.data_classes import data_modeling as dm

from cognite.pygen import config as pygen_config
from cognite.pygen.utils.text import create_name

from .data_classes import DataClass, EdgeDataClass, NodeDataClass
from .fields import BaseConnectionField, CDFExternalField
from .filter_methods import FilterMethod, FilterParameter

if TYPE_CHECKING:
    from cognite.pygen._core.generators import APIGenerator


@dataclass(frozen=True)
class APIClass:
    """This is the base class for all API classes.

    Args:
        parent_attribute: The name of the attribute in the client class that this API class is attached.
        name: The name of the API class.
        file_name: The name of the file where the API class is located.

    """

    parent_attribute: str
    name: str
    file_name: str


@dataclass(frozen=True)
class NodeAPIClass(APIClass):
    is_edge_class: bool
    view_id: dm.ViewId

    @classmethod
    def from_view(
        cls, view_id: dm.ViewId, base_name: str, is_edge_class: bool, api_class: pygen_config.APIClassNaming
    ) -> NodeAPIClass:
        file_name = create_name(base_name, api_class.file_name)
        class_name = create_name(base_name, api_class.name)
        return cls(
            parent_attribute=create_name(base_name, api_class.client_attribute),
            name=f"{class_name}API",
            file_name=file_name,
            is_edge_class=is_edge_class,
            view_id=view_id,
        )


@dataclass(frozen=True)
class QueryAPIClass(APIClass):
    data_class: DataClass

    @classmethod
    def create(cls, data_class: DataClass, base_name: str, api_class: pygen_config.APIClassNaming) -> QueryAPIClass:
        file_name = create_name(base_name, api_class.file_name)
        class_name = create_name(base_name, api_class.name)
        parent_attribute = create_name(base_name, api_class.client_attribute)
        return cls(
            parent_attribute=f"{parent_attribute}_query",  # Not used.
            name=f"{class_name}QueryAPI",
            file_name=f"{file_name}_query",
            data_class=data_class,
        )


@dataclass(frozen=True)
class EdgeAPIClass(APIClass):
    start_class: DataClass
    end_class: DataClass
    edge_class: DataClass | None
    field_name: str
    type: dm.DirectRelationReference
    direction: Literal["outwards", "inwards"]
    filter_method: FilterMethod
    doc_name: str
    query: QueryAPIClass
    end_view_id: dm.ViewId
    end_filter_method: FilterMethod
    field: BaseConnectionField
    has_default_instance_space: bool

    @property
    def has_edge_class(self) -> bool:
        return self.edge_class is not None

    def filter_parameters(
        self,
        include_nodes: bool = True,
        include_end_node: bool = False,
        suffix_edge: bool = False,
        case: Literal["signature", "docs", "filter_call"] = "signature",
    ) -> Iterator[FilterParameter]:
        if include_nodes:
            nodes: dict[str, FilterParameter] = {}
            type_ = "dm.NodeId | list[dm.NodeId]"
            if self.has_default_instance_space:
                type_ = f"str | list[str] | {type_}"
            nodes["from"] = FilterParameter(
                name=f"from_{self.start_class.variable}",
                type_=type_,
                description=f"ID of the source { self.start_class.doc_name}.",
                default=None,
            )
            if self.has_default_instance_space:
                nodes["from_space"] = FilterParameter(
                    name=f"from_{self.start_class.variable}_space",
                    type_="str",
                    description=f"Location of the {self.start_class.doc_list_name}.",
                    default="DEFAULT_INSTANCE_SPACE",
                    is_nullable=False,
                )
            nodes["to"] = FilterParameter(
                name=f"to_{self.end_class.variable}",
                type_=type_,
                description=f"ID of the target { self.end_class.doc_name}.",
                default=None,
            )
            if self.has_default_instance_space:
                nodes["to_space"] = FilterParameter(
                    name=f"to_{self.end_class.variable}_space",
                    type_="str",
                    description=f"Location of the {self.end_class.doc_list_name}.",
                    default="DEFAULT_INSTANCE_SPACE",
                    is_nullable=False,
                )

            if case == "filter_call" and self.direction == "inwards":
                if self.has_default_instance_space:
                    yield from [nodes["to"], nodes["to_space"], nodes["from"], nodes["from_space"]]
                else:
                    yield from [nodes["to"], nodes["from"]]
            else:
                yield from nodes.values()

        if include_end_node:
            yield from self.end_filter_method.parameters
        if suffix_edge:
            yield from [param.copy(name=f"{param.name}_edge") for param in self.filter_method.parameters]
        else:
            yield from self.filter_method.parameters

    @classmethod
    def from_fields(
        cls,
        field: BaseConnectionField,
        data_class: DataClass,
        base_name: str,
        query_class_by_view_id: dict[dm.ViewId, QueryAPIClass],
        api_generator_by_view_id: dict[dm.ViewId, APIGenerator],
        has_default_instance_space: bool,
        pygen_config: pygen_config.PygenConfig,
    ) -> EdgeAPIClass:
        if field.is_direct_relation:
            raise ValueError("Bug in Pygen: Expected only Edge Connection Fields")
        api_class = pygen_config.naming.api_class
        base_name = f"{base_name}_{field.name}"
        file_name = create_name(base_name, api_class.file_name)
        class_name = create_name(base_name, api_class.name)
        parent_attribute = create_name(field.name, api_class.client_attribute)

        # This is always true for Edge Connection Fields
        end_class = cast(NodeDataClass, field.destination_class)

        edge_class: EdgeDataClass | None = None
        if field.edge_class:
            edge_class = field.edge_class
            filter_method = FilterMethod.from_fields(
                edge_class.fields, pygen_config.filtering, has_default_instance_space, is_edge_class=True
            )
        else:  #  NodeDataClass
            filter_method = FilterMethod.from_fields(
                [], pygen_config.filtering, has_default_instance_space, is_edge_class=False
            )

        return cls(
            parent_attribute=f"{parent_attribute}_edge",
            name=f"{class_name}API",
            file_name=file_name,
            edge_class=edge_class,
            field_name=field.name,
            # This is always true for Edge Connection Fields
            type=field.edge_type,  # type: ignore[arg-type]
            start_class=data_class,
            end_class=end_class,
            filter_method=filter_method,
            doc_name=create_name(field.name, api_class.doc_name, python_variable=False),
            query=query_class_by_view_id[end_class.view_id],
            direction=field.edge_direction,
            end_view_id=end_class.view_id,
            end_filter_method=api_generator_by_view_id[end_class.view_id].list_method,
            field=field,
            has_default_instance_space=has_default_instance_space,
        )


@dataclass(frozen=True)
class TimeSeriesAPIClass(APIClass):
    query_class: str
    prop_name: str
    variable: str

    @classmethod
    def from_field(
        cls, field: CDFExternalField, base_name: str, api_class: pygen_config.APIClassNaming
    ) -> TimeSeriesAPIClass:
        base_name = f"{base_name}_{field.name}"
        file_name = create_name(base_name, api_class.file_name)
        class_name = create_name(base_name, api_class.name)
        parent_attribute = create_name(field.name, api_class.client_attribute)
        variable = create_name(field.name, api_class.variable)

        return cls(
            parent_attribute=f"{parent_attribute}",
            name=f"{class_name}API",
            file_name=f"{file_name}",
            query_class=f"{class_name}Query",
            prop_name=field.prop_name,
            variable=variable,
        )


@dataclass(frozen=True)
class MultiAPIClass:
    """This represents a set of APIs which are generated from a single data model.

    The motivation for having this class is the case when you want to create one SDK for multiple data models.

    Args:
        sub_apis_by_view_id: A dictionary mapping view IDs to the corresponding API classes.
        parent_attribute: The name of the attribute in the client class that this API class is attached
        name: The name of the API class.
        model: The data model that the APIs are generated from.
    """

    sub_apis_by_view_id: dict[dm.ViewId, APIClass]
    parent_attribute: str
    name: str
    model: dm.DataModel[dm.View]

    @property
    def model_id(self) -> dm.DataModelId:
        return self.model.as_id()

    @classmethod
    def from_data_model(
        cls,
        data_model: dm.DataModel[dm.View],
        api_class_by_view_id: dict[dm.ViewId, APIClass],
        multi_api_class: pygen_config.MultiAPIClassNaming,
    ) -> MultiAPIClass:
        sub_apis: dict[dm.ViewId, APIClass] = {}
        for view in sorted(data_model.views, key=lambda v: (v.name or v.external_id, v.space, v.version)):
            if view.as_id() in api_class_by_view_id:
                sub_apis[view.as_id()] = api_class_by_view_id[view.as_id()]

        data_model_name = data_model.name or data_model.external_id

        return cls(
            sub_apis_by_view_id=sub_apis,
            parent_attribute=create_name(data_model_name, multi_api_class.client_attribute),
            name=f"{create_name(data_model_name, multi_api_class.name)}APIs",
            model=data_model,
        )



import datetime
import math
import time
import warnings
from collections import defaultdict
from collections.abc import Collection, Iterable, Iterator, MutableSequence, Sequence
from contextlib import suppress
from dataclasses import dataclass
from typing import (
    Any,
    Literal,
    SupportsIndex,
    cast,
    overload,
)

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes._base import CogniteObject
from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling.instances import Instance
from cognite.client.exceptions import CogniteAPIError

DEFAULT_QUERY_LIMIT = 5
INSTANCE_QUERY_LIMIT = 1_000
# The limit used for the In filter in /search
IN_FILTER_CHUNK_SIZE = 100
# This is the actual limit of the API, we typically set it to a lower value to avoid hitting the limit.
# The actual instance query limit is 10_000, but we set it to 5_000 such that is matches the In filter
# which we use in /search for reverse of list direct relations.
ACTUAL_INSTANCE_QUERY_LIMIT = 5_000
# The minimum estimated seconds before print progress on a query
MINIMUM_ESTIMATED_SECONDS_BEFORE_PRINT_PROGRESS = 30
PRINT_PROGRESS_PER_N_NODES = 10_000
SEARCH_LIMIT = 1_000
AGGREGATION_LIMIT = 1_000


@dataclass(frozen=True)
class ViewPropertyId(CogniteObject):
    view: dm.ViewId
    property: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> "ViewPropertyId":
        return cls(
            view=dm.ViewId.load(resource["view"]),
            property=resource["identifier"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "view": self.view.dump(camel_case=camel_case, include_type=False),
            "identifier": self.property,
        }


class _NotSetSentinel:
    """This is a special class that indicates that a value has not been set.
    It is used when we need to distinguish between not set and None."""

    ...


class QueryReducingBatchSize(UserWarning):
    """Raised when a query is too large and the batch size must be reduced."""

    ...


def chunker(sequence: Sequence, chunk_size: int) -> Iterator[Sequence]:
    """
    Split a sequence into chunks of size chunk_size.

    Args:
        sequence: The sequence to split.
        chunk_size: The size of each chunk.

    Returns:
        An iterator over the chunks.

    """
    for i in range(0, len(sequence), chunk_size):
        yield sequence[i : i + chunk_size]


class QueryStep:
    def __init__(
        self,
        name: str,
        expression: dm.query.ResultSetExpression,
        view_id: dm.ViewId | None = None,
        max_retrieve_limit: int = -1,
        select: dm.query.Select | None | type[_NotSetSentinel] = _NotSetSentinel,
        raw_filter: dm.Filter | None = None,
        connection_type: Literal["reverse-list"] | None = None,
        view_property: ViewPropertyId | None = None,
        selected_properties: list[str] | None = None,
    ):
        self.name = name
        self.expression = expression
        self.view_id = view_id
        self.max_retrieve_limit = max_retrieve_limit
        self.select: dm.query.Select | None
        if select is _NotSetSentinel:
            try:
                self.select = self._default_select()
            except NotImplementedError:
                raise ValueError(f"You need to provide a select to instantiate a {type(self).__name__}") from None
        else:
            self.select = select  # type: ignore[assignment]
        self.raw_filter = raw_filter
        self.connection_type = connection_type
        self.view_property = view_property
        self.selected_properties = selected_properties
        self._max_retrieve_batch_limit = ACTUAL_INSTANCE_QUERY_LIMIT
        self.cursor: str | None = None
        self.total_retrieved: int = 0
        self.last_batch_count: int = 0
        self.results: list[Instance] = []

    def _default_select(self) -> dm.query.Select:
        if self.view_id is None:
            return dm.query.Select()
        else:
            return dm.query.Select([dm.query.SourceSelector(self.view_id, ["*"])])

    @property
    def is_queryable(self) -> bool:
        # We cannot query across reverse-list connections
        return self.connection_type != "reverse-list"

    @property
    def from_(self) -> str | None:
        return self.expression.from_

    @property
    def is_single_direct_relation(self) -> bool:
        return isinstance(self.expression, dm.query.NodeResultSetExpression) and self.expression.through is not None

    @property
    def node_expression(self) -> dm.query.NodeResultSetExpression | None:
        if isinstance(self.expression, dm.query.NodeResultSetExpression):
            return self.expression
        return None

    @property
    def edge_expression(self) -> dm.query.EdgeResultSetExpression | None:
        if isinstance(self.expression, dm.query.EdgeResultSetExpression):
            return self.expression
        return None

    @property
    def node_results(self) -> Iterable[dm.Node]:
        return (item for item in self.results if isinstance(item, dm.Node))

    @property
    def edge_results(self) -> Iterable[dm.Edge]:
        return (item for item in self.results if isinstance(item, dm.Edge))

    def update_expression_limit(self) -> None:
        if self.is_unlimited:
            self.expression.limit = self._max_retrieve_batch_limit
        else:
            self.expression.limit = max(min(INSTANCE_QUERY_LIMIT, self.max_retrieve_limit - self.total_retrieved), 0)

    def reduce_max_batch_limit(self) -> bool:
        self._max_retrieve_batch_limit = max(1, self._max_retrieve_batch_limit // 2)
        return self._max_retrieve_batch_limit > 1

    @property
    def is_unlimited(self) -> bool:
        return self.max_retrieve_limit in {None, -1, math.inf}

    @property
    def is_finished(self) -> bool:
        return (
            (not self.is_unlimited and self.total_retrieved >= self.max_retrieve_limit)
            or self.cursor is None
            or self.last_batch_count == 0
            # Single direct relations are dependent on the parent node,
            # so we assume that the parent node is the limiting factor.
            or self.is_single_direct_relation
        )

    def count_total(self, cognite_client: CogniteClient) -> float:
        if self.view_id is None:
            raise ValueError("Cannot count total if select is not set")

        return cognite_client.data_modeling.instances.aggregate(
            self.view_id, Count("externalId"), filter=self.raw_filter
        ).value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, from={self.from_!r}, results={len(self.results)})"


class QueryBuilder(list, MutableSequence[QueryStep]):
    """This is a helper class to build and execute a query. It is responsible for
    doing the paging of the query and keeping track of the results."""

    def __init__(self, steps: Collection[QueryStep] | None = None):
        super().__init__(steps or [])

    def _reset(self):
        for expression in self:
            expression.total_retrieved = 0
            expression.cursor = None
            expression.results = []

    def _update_expression_limits(self) -> None:
        for expression in self:
            expression.update_expression_limit()

    def _build(self) -> tuple[dm.query.Query, list[QueryStep], set[str]]:
        with_ = {step.name: step.expression for step in self if step.is_queryable}
        select = {step.name: step.select for step in self if step.select is not None and step.is_queryable}
        cursors = self._cursors

        step_by_name = {step.name: step for step in self}
        search: list[QueryStep] = []
        temporary_select: set[str] = set()
        for step in self:
            if step.is_queryable:
                continue
            if step.node_expression is not None:
                search.append(step)
                # Ensure that select is set for the parent
                if step.from_ in select or step.from_ is None:
                    continue
                view_id = step_by_name[step.from_].view_id
                if view_id is None:
                    continue
                select[step.from_] = dm.query.Select([dm.query.SourceSelector(view_id, ["*"])])
                temporary_select.add(step.from_)
        return dm.query.Query(with_=with_, select=select, cursors=cursors), search, temporary_select

    def _dump_yaml(self) -> str:
        return self._build()[0].dump_yaml()

    @property
    def _cursors(self) -> dict[str, str | None]:
        return {expression.name: expression.cursor for expression in self if expression.is_queryable}

    def _update(self, batch: dm.query.QueryResult):
        for expression in self:
            if expression.name not in batch:
                continue
            expression.last_batch_count = len(batch[expression.name])
            expression.total_retrieved += expression.last_batch_count
            expression.cursor = batch.cursors.get(expression.name)
            expression.results.extend(batch[expression.name].data)

    @property
    def _is_finished(self) -> bool:
        return self[0].is_finished

    def _reduce_max_batch_limit(self) -> bool:
        for expression in self:
            if not expression.reduce_max_batch_limit():
                return False
        return True

    def execute_query(self, client: CogniteClient, remove_not_connected: bool = False) -> dict[str, list[Instance]]:
        self._reset()
        query, to_search, temp_select = self._build()

        if not self:
            raise ValueError("No query steps to execute")

        count: float | None = None
        with suppress(ValueError, CogniteAPIError):
            count = self[0].count_total(client)

        is_large_query = False
        last_progress_print = 0
        nodes_per_second = 0.0
        while True:
            self._update_expression_limits()
            query.cursors = self._cursors
            t0 = time.time()
            try:
                batch = client.data_modeling.instances.query(query)
            except CogniteAPIError as e:
                if e.code == 408:
                    # Too big query, try to reduce the limit
                    if self._reduce_max_batch_limit():
                        continue
                    new_limit = self[0]._max_retrieve_batch_limit
                    warnings.warn(
                        f"Query is too large, reducing batch size to {new_limit:,}, and trying again",
                        QueryReducingBatchSize,
                        stacklevel=2,
                    )

                raise e

            self._fetch_reverse_direct_relation_of_lists(client, to_search, batch)

            for name in temp_select:
                batch.pop(name, None)

            last_execution_time = time.time() - t0

            self._update(batch)
            if self._is_finished:
                break

            if count is None:
                continue
            # Estimate the number of nodes per second using exponential moving average
            last_batch_nodes_per_second = len(batch[self[0].name]) / last_execution_time
            if nodes_per_second == 0.0:
                nodes_per_second = last_batch_nodes_per_second
            else:
                nodes_per_second = 0.1 * last_batch_nodes_per_second + 0.9 * nodes_per_second
            # Estimate the time to completion
            remaining_nodes = count - self[0].total_retrieved
            remaining_time = remaining_nodes / nodes_per_second

            if is_large_query and (self[0].total_retrieved - last_progress_print) > PRINT_PROGRESS_PER_N_NODES:
                estimate = datetime.timedelta(seconds=round(remaining_time, 0))
                print(
                    f"Progress: {self[0].total_retrieved:,}/{count:,} nodes retrieved. "
                    f"Estimated time to completion: {estimate}"
                )
                last_progress_print = self[0].total_retrieved

            if is_large_query is False and remaining_time > MINIMUM_ESTIMATED_SECONDS_BEFORE_PRINT_PROGRESS:
                is_large_query = True
                print("Large query detected. Will print progress.")

        if remove_not_connected and len(self) > 1:
            _QueryResultCleaner(self).clean()

        return {step.name: step.results for step in self}

    @staticmethod
    def _fetch_reverse_direct_relation_of_lists(
        client: CogniteClient, to_search: list[QueryStep], batch: dm.query.QueryResult
    ) -> None:
        """Reverse direct relations for lists are not supported by the query API.
        This method fetches them separately."""
        for step in to_search:
            if step.from_ is None or step.from_ not in batch:
                continue
            item_ids = [node.as_id() for node in batch[step.from_].data]
            if not item_ids:
                continue

            view_id = step.view_id
            expression = step.node_expression
            if view_id is None or expression is None:
                raise ValueError(
                    "Invalid state of the query. Search should always be a node expression with view properties"
                )
            if expression.through is None:
                raise ValueError("Missing through set in a reverse-list query")
            limit = SEARCH_LIMIT if step.is_unlimited else min(step.max_retrieve_limit, SEARCH_LIMIT)

            step_result = dm.NodeList[dm.Node]([])
            for item_ids_chunk in chunker(item_ids, IN_FILTER_CHUNK_SIZE):
                is_items = dm.filters.In(view_id.as_property_ref(expression.through.property), item_ids_chunk)
                is_selected = is_items if step.raw_filter is None else dm.filters.And(is_items, step.raw_filter)

                chunk_result = client.data_modeling.instances.search(
                    view_id, properties=None, filter=is_selected, limit=limit
                )
                step_result.extend(chunk_result)

            batch[step.name] = dm.NodeListWithCursor(step_result, None)
        return None

    def get_from(self) -> str | None:
        if len(self) == 0:
            return None
        return self[-1].name

    def create_name(self, from_: str | None) -> str:
        if from_ is None:
            return "0"
        return f"{from_}_{len(self)}"

    def append(self, __object: QueryStep, /) -> None:
        # Extra validation to ensure all assumptions are met
        if len(self) == 0:
            if __object.from_ is not None:
                raise ValueError("The first step should not have a 'from_' value")
        else:
            if __object.from_ is None:
                raise ValueError("The 'from_' value should be set")
        super().append(__object)

    def extend(self, __iterable: Iterable[QueryStep], /) -> None:
        for item in __iterable:
            self.append(item)

    # The implementations below are to get proper type hints
    def __iter__(self) -> Iterator[QueryStep]:
        return super().__iter__()

    @overload
    def __getitem__(self, item: SupportsIndex) -> QueryStep: ...

    @overload
    def __getitem__(self, item: slice) -> "QueryBuilder": ...

    def __getitem__(self, item: SupportsIndex | slice) -> "QueryStep | QueryBuilder":
        value = super().__getitem__(item)
        if isinstance(item, slice):
            return QueryBuilder(value)  # type: ignore[arg-type]
        return cast(QueryStep, value)


class _QueryResultCleaner:
    """Remove nodes and edges that are not connected through the entire query"""

    def __init__(self, steps: list[QueryStep]):
        self._tree = self._create_tree(steps)
        self._root = steps[0]

    @classmethod
    def _create_tree(cls, steps: list[QueryStep]) -> dict[str, list[QueryStep]]:
        tree: dict[str, list[QueryStep]] = defaultdict(list)
        for step in steps:
            if step.from_ is None:
                continue
            tree[step.from_].append(step)
        return dict(tree)

    def clean(self) -> None:
        self._clean(self._root)

    @staticmethod
    def as_node_id(direct_relation: dm.DirectRelationReference | dict[str, str]) -> dm.NodeId:
        if isinstance(direct_relation, dict):
            return dm.NodeId(direct_relation["space"], direct_relation["externalId"])

        return dm.NodeId(direct_relation.space, direct_relation.external_id)

    def _clean(self, step: QueryStep) -> tuple[set[dm.NodeId], str | None]:
        if step.name not in self._tree:
            # Leaf Node
            direct_relation: str | None = None
            if step.node_expression and (through := step.node_expression.through) is not None:
                direct_relation = through.property
                if step.node_expression.direction == "inwards":
                    return {
                        node_id for item in step.node_results for node_id in self._get_relations(item, direct_relation)
                    }, None

            return {item.as_id() for item in step.results}, direct_relation  # type: ignore[attr-defined]

        expected_ids_by_property: dict[str | None, set[dm.NodeId]] = {}
        for child in self._tree[step.name]:
            child_ids, property_id = self._clean(child)
            if property_id not in expected_ids_by_property:
                expected_ids_by_property[property_id] = child_ids
            else:
                expected_ids_by_property[property_id] |= child_ids

        if step.node_expression is not None:
            filtered_results: list[Instance] = []
            for node in step.node_results:
                if self._is_connected_node(node, expected_ids_by_property):
                    filtered_results.append(node)
            step.results = filtered_results
            direct_relation = None if step.node_expression.through is None else step.node_expression.through.property
            return {node.as_id() for node in step.node_results}, direct_relation

        if step.edge_expression:
            if len(expected_ids_by_property) > 1 or None not in expected_ids_by_property:
                raise RuntimeError(f"Invalid state of {type(self).__name__}")
            expected_ids = expected_ids_by_property[None]
            if step.edge_expression.direction == "outwards":
                step.results = [edge for edge in step.edge_results if self.as_node_id(edge.end_node) in expected_ids]
                return {self.as_node_id(edge.start_node) for edge in step.edge_results}, None
            else:  # inwards
                step.results = [edge for edge in step.edge_results if self.as_node_id(edge.start_node) in expected_ids]
                return {self.as_node_id(edge.end_node) for edge in step.edge_results}, None

        raise TypeError(f"Unsupported query step type: {type(step)}")

    @classmethod
    def _is_connected_node(cls, node: dm.Node, expected_ids_by_property: dict[str | None, set[dm.NodeId]]) -> bool:
        if not expected_ids_by_property:
            return True
        if None in expected_ids_by_property:
            if node.as_id() in expected_ids_by_property[None]:
                return True
            if len(expected_ids_by_property) == 1:
                return False
        node_properties = next(iter(node.properties.values()))
        for property_id, expected_ids in expected_ids_by_property.items():
            if property_id is None:
                continue
            value = node_properties.get(property_id)
            if value is None:
                continue
            elif isinstance(value, list):
                if {cls.as_node_id(item) for item in value if isinstance(item, dict)} & expected_ids:
                    return True
            elif isinstance(value, dict) and cls.as_node_id(value) in expected_ids:
                return True
        return False

    @classmethod
    def _get_relations(cls, node: dm.Node, property_id: str) -> Iterable[dm.NodeId]:
        if property_id is None:
            return {node.as_id()}
        value = next(iter(node.properties.values())).get(property_id)
        if isinstance(value, list):
            return [cls.as_node_id(item) for item in value if isinstance(item, dict)]
        elif isinstance(value, dict):
            return [cls.as_node_id(value)]
        return []



"""This module contains the base class for all fields."""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, TypeVar, cast

from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes.data_modeling.views import (
    ViewProperty,
)

from cognite.pygen import config as pygen_config
from cognite.pygen.config.reserved_words import is_reserved_word
from cognite.pygen.utils.text import create_name, to_words

if TYPE_CHECKING:
    from cognite.pygen._core.models.data_classes import EdgeDataClass, NodeDataClass

_PRIMITIVE_TYPES = (dm.Text, dm.Boolean, dm.Float32, dm.Float64, dm.Int32, dm.Int64, dm.Timestamp, dm.Date, dm.Json)
_EXTERNAL_TYPES = (dm.TimeSeriesReference, dm.FileReference, dm.SequenceReference)


@dataclass(frozen=True)
class Field:
    """
    A field represents a pydantic field in the generated pydantic class.

    Args:
        name: The name of the field. This is used in the generated Python code.
        doc_name: The name of the field in the documentation.
        prop_name: The name of the property in the data model. This is used when reading and writing to CDF.
        pydantic_field: The name to use for the import 'from pydantic import Field'. This is used in the edge case
                        when the name 'Field' name clashes with the data model class name.

    """

    name: str
    doc_name: str
    prop_name: str
    description: str | None
    pydantic_field: Literal["Field", "pydantic.Field"]

    @property
    def need_alias(self) -> bool:
        return self.name != self.prop_name

    @classmethod
    def from_property(
        cls,
        prop_id: str,
        prop: ViewProperty,
        node_class_by_view_id: dict[dm.ViewId, NodeDataClass],
        edge_class_by_view_id: dict[dm.ViewId, EdgeDataClass],
        config: pygen_config.PygenConfig,
        view_id: dm.ViewId,
        pydantic_field: Literal["Field", "pydantic.Field"],
        has_default_instance_space: bool,
        direct_relations_by_view_id: dict[dm.ViewId, set[str]],
        view_by_id: dict[dm.ViewId, dm.View],
    ) -> Field | None:
        from .cdf_reference import CDFExternalField
        from .connections import BaseConnectionField
        from .primitive import BasePrimitiveField

        field_naming = config.naming.field
        name = create_name(prop_id, field_naming.name)
        if is_reserved_word(name, "field", view_id, prop_id):
            name = f"{name}_"

        doc_name = to_words(name, singularize=True)
        variable = create_name(prop_id, field_naming.variable)
        description: str | None = None
        if hasattr(prop, "description") and isinstance(prop.description, str):
            # This is a workaround for the fact that the description can contain curly quotes
            # which is ruff will complain about. (These comes from che Core model)
            description = prop.description.replace("‘", "'").replace("’", "'")  # noqa: RUF001

        base = cls(
            name=name,
            doc_name=doc_name,
            prop_name=prop_id,
            description=description,
            pydantic_field=pydantic_field,
        )
        if isinstance(prop, dm.ConnectionDefinition) or (
            isinstance(prop, dm.MappedProperty) and isinstance(prop.type, dm.DirectRelation)
        ):
            return BaseConnectionField.load(
                base,
                prop,
                variable,
                node_class_by_view_id,
                edge_class_by_view_id,
                has_default_instance_space,
                view_id,
                direct_relations_by_view_id,
            )
        elif isinstance(prop, dm.MappedProperty) and isinstance(prop.type, dm.CDFExternalIdReference):
            return CDFExternalField.load(base, prop, variable)
        elif isinstance(prop, dm.MappedProperty) and isinstance(prop.type, dm.PropertyType):
            return BasePrimitiveField.load(base, prop, variable)
        else:
            warnings.warn(
                f"Unsupported property type: {type(prop)}. Skipping field {prop_id}", UserWarning, stacklevel=2
            )
            return None

    def as_read_type_hint(self) -> str:
        raise NotImplementedError()

    def as_write_type_hint(self) -> str:
        raise NotImplementedError()

    def as_graphql_type_hint(self) -> str:
        raise NotImplementedError()

    def as_typed_hint(self, operation: Literal["write", "read"] = "write") -> str:
        raise NotImplementedError()

    def as_typed_init_set(self) -> str:
        return self.name

    def as_write(self) -> str:
        """Used in the .as_write() method for the read version of the data class."""
        raise NotImplementedError

    def as_write_graphql(self) -> str:
        """Used in the .as_write() method for the graphQL version of the data class."""
        return self.as_write()

    def as_read_graphql(self) -> str:
        """Used in the .as_read() method for the graphQL version of the data class."""
        raise NotImplementedError

    def as_value(self) -> str:
        """Used in the ._to_instances_write() method to write the value of the field to the node instance.
        This should only be implemented for container fields, i.e., fields that store their value in a container."""
        raise NotImplementedError

    @property
    def argument_documentation(self) -> str:
        if self.description:
            return self.description
        else:
            return f"The {self.doc_name} field."

    # The properties below are overwritten in the child classes
    @property
    def is_connection(self) -> bool:
        return False

    @property
    def is_time_field(self) -> bool:
        return False

    @property
    def is_timestamp(self) -> bool:
        return False

    @property
    def is_time_series(self) -> bool:
        return False

    @property
    def is_list(self) -> bool:
        return False

    @property
    def is_text_field(self) -> bool:
        return False

    @property
    def is_write_field(self) -> bool:
        return True


T_Field = TypeVar("T_Field", bound=Field)


def _get_prop_view_id(prop_id: str, view_id: dm.ViewId, view_by_id: dict[dm.ViewId, dm.View]) -> dm.ViewId:
    return cast(dm.ViewId, _search_source_view(prop_id, view_by_id[view_id], view_by_id))


def _search_source_view(prop_id: str, view: dm.View, view_by_id: dict[dm.ViewId, dm.View]) -> dm.ViewId | None:
    if prop_id in view.properties and not view.implements:
        return view.as_id()
    elif prop_id not in view.properties:
        return None
    # In properties, and have parents.
    candidates = set()
    for parent_id in view.implements:
        parent = view_by_id[parent_id]
        parent_view_id = _search_source_view(prop_id, parent, view_by_id)
        if parent_view_id:
            candidates.add(parent_view_id)
    if len(candidates) == 1:
        return candidates.pop()
    elif len(candidates) > 1:
        used = candidates.pop()
        warnings.warn(
            f"Multiple candidates for source view of property {prop_id}. Using {used}",
            UserWarning,
            stacklevel=2,
        )
        return used
    else:
        # Original to this view.
        return view.as_id()



"""This module contains the fields that are references to CDF External Fields. These fields are referencing
either TimeSeries, Sequence, or File objects. In other words, these fields are references to objects that are
outside of the data model."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from cognite.client.data_classes import data_modeling as dm
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType

from .base import Field
from .primitive import BasePrimitiveField, PrimitiveListField


@dataclass(frozen=True)
class CDFExternalField(BasePrimitiveField):
    """This represents a field that is a reference to a CDF External Field.

    For example, a field that is a reference to a TimeSeries, Sequence, or File.
    """

    @property
    def type_as_string(self) -> str:
        if isinstance(self.type_, dm.TimeSeriesReference):
            return "TimeSeries"
        elif isinstance(self.type_, dm.SequenceReference):
            return "SequenceRead"
        elif isinstance(self.type_, dm.FileReference):
            return "FileMetadata"
        else:
            raise ValueError(f"Unknown CDF External Field type: {self.type_}")

    @property
    def resource_write_name(self) -> str:
        if isinstance(self.type_, dm.TimeSeriesReference):
            return "time_series"
        elif isinstance(self.type_, dm.SequenceReference):
            return "sequences"
        elif isinstance(self.type_, dm.FileReference):
            return "files"
        else:
            raise ValueError(f"Unknown CDF External Field type: {self.type_}")

    def cognite_type_name(self, format: Literal["read", "write"]) -> str:
        type_name = self.type_name
        if format == "write":
            return f"Cognite{type_name}Write"
        else:
            return f"Cognite{type_name}"

    @property
    def type_name(self) -> str:
        return self.type_as_string.removesuffix("Read")

    @property
    def is_time_series(self) -> bool:
        return isinstance(self.type_, dm.TimeSeriesReference)

    @property
    def is_list(self) -> bool:
        return isinstance(self.type_, ListablePropertyType) and self.type_.is_list

    def as_read_type_hint(self) -> str:
        return self._create_type_hint(self.type_as_string)

    def as_graphql_type_hint(self) -> str:
        type_ = f"{self.type_name}GraphQL"
        if self.need_alias:
            return f'Optional[{type_}] = {self.pydantic_field}(None, alias="{self.prop_name}")'
        else:
            return f"Optional[{type_}] = None"

    def as_write_type_hint(self) -> str:
        type_ = f"{self.type_name}Write"
        return self._create_type_hint(type_)

    def _create_type_hint(self, type_: str) -> str:
        # CDF External Fields are always nullable
        if self.need_alias:
            return f'Union[{type_}, str, None] = {self.pydantic_field}(None, alias="{self.prop_name}")'
        else:
            return f"Union[{type_}, str, None] = None"

    def as_write(self) -> str:
        type_ = self.type_name
        return f"self.{self.name}.as_write() if isinstance(self.{self.name}, Cognite{type_}) else self.{self.name}"

    def as_write_graphql(self) -> str:
        return f"self.{self.name}.as_write() if self.{self.name} else None"

    def as_read_graphql(self) -> str:
        # Read is only used in graphql
        return f"self.{self.name}.as_read() if self.{self.name} else None"

    def as_value(self) -> str:
        return (
            f"self.{self.name} if isinstance(self.{self.name}, str) or self.{self.name} is None "
            f"else self.{self.name}.external_id"
        )

    @classmethod
    def load(cls, base: Field, prop: dm.MappedProperty, variable: str) -> CDFExternalField | None:
        if not isinstance(prop.type, dm.CDFExternalIdReference):
            return None
        if prop.type.is_list:
            return CDFExternalListField(
                name=base.name,
                prop_name=base.prop_name,
                doc_name=base.doc_name,
                type_=prop.type,
                is_nullable=prop.nullable,
                description=prop.description,
                pydantic_field=base.pydantic_field,
                variable=variable,
            )
        else:
            return CDFExternalField(
                name=base.name,
                prop_name=base.prop_name,
                doc_name=base.doc_name,
                type_=prop.type,
                is_nullable=prop.nullable,
                description=prop.description,
                pydantic_field=base.pydantic_field,
            )


@dataclass(frozen=True)
class CDFExternalListField(PrimitiveListField, CDFExternalField):
    """
    This represents a list of CDF types such as list[TimeSeries], list[Sequence], or list[File].
    """

    def as_read_type_hint(self) -> str:
        return self._create_type_hint(self.type_as_string)

    def _create_type_hint(self, type_: str) -> str:
        if self.need_alias:
            return f'Optional[list[Union[{type_}, str]]] = {self.pydantic_field}(None, alias="{self.prop_name}")'
        else:
            return f"Optional[list[Union[{type_}, str]]] = None"

    def as_write_type_hint(self) -> str:
        type_ = f"{self.type_name}Write"
        return self._create_type_hint(type_)

    def as_graphql_type_hint(self) -> str:
        type_ = f"{self.type_name}GraphQL"
        if self.need_alias:
            return f'Optional[list[{type_}]] = {self.pydantic_field}(None, alias="{self.prop_name}")'
        else:
            return f"Optional[list[{type_}]] = None"

    def as_write(self) -> str:
        return (
            f"[{self.variable}.as_write() if isinstance({self.variable}, {self.cognite_type_name('read')}) "
            f"else {self.variable}"
            f" for {self.variable} in self.{self.name}] if self.{self.name} is not None else None"
        )

    def as_write_graphql(self) -> str:
        return (
            f"[{self.variable}.as_write() for {self.variable} in self.{self.name} or []]"
            f" if self.{self.name} is not None else None"
        )

    def as_read_graphql(self) -> str:
        # Read is only used in graphql
        return (
            f"[{self.variable}.as_read() for {self.variable} in self.{self.name} or []]"
            f" if self.{self.name} is not None else None"
        )

    def as_value(self) -> str:
        return (
            f"[{self.variable} if isinstance({self.variable}, str) else {self.variable}.external_id "
            f"for {self.variable} in self.{self.name} or []] if self.{self.name} is not None else None"
        )



import copy
import itertools
import warnings
from collections import defaultdict
from collections.abc import Callable, Sequence
from functools import cached_property
from typing import Any, Literal, TypeAlias, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import filters
from cognite.client.data_classes.aggregations import Aggregation
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList
from cognite.client.data_classes.data_modeling.views import ReverseDirectRelation, ViewProperty
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite.pygen._version import __version__

from .query_builder import (
    AGGREGATION_LIMIT,
    IN_FILTER_CHUNK_SIZE,
    SEARCH_LIMIT,
    QueryBuilder,
    QueryStep,
    ViewPropertyId,
    chunker,
)

_NODE_PROPERTIES = frozenset(
    {"externalId", "space", "version", "lastUpdatedTime", "createdTime", "deletedTime", "type"}
)

Properties: TypeAlias = list[str | dict[str, list[str | dict[str, Any]]]]


class QueryExecutor:
    def __init__(self, client: CogniteClient, views: Sequence[dm.View] | None = None):
        self._client = client
        # Used for aggregated logging of requests
        client.config.client_name = f"CognitePygen:{__version__}:QueryExecutor"
        self._view_by_id: dict[dm.ViewId, dm.View] = {view.as_id(): view for view in views or []}

    def _get_view(self, view_id: dm.ViewId) -> dm.View:
        if view_id not in self._view_by_id:
            view = self._client.data_modeling.views.retrieve(view_id, all_versions=False)
            if not view:
                raise CogniteAPIError(f"View not found: {view_id!r}", code=200)
            self._view_by_id[view_id] = view[0]
        return self._view_by_id[view_id]

    @staticmethod
    def _as_property_list(properties: Properties, operation: str) -> list[str]:
        output = []
        is_nested_supported = operation == "list"
        for prop in properties:
            if isinstance(prop, str):
                output.append(prop)
            elif isinstance(prop, dict) and is_nested_supported:
                if len(prop) != 1:
                    raise ValueError(f"Unexpected nested property: {prop}")
                key = next(iter(prop.keys()))
                output.append(key)
            elif isinstance(prop, dict):
                raise ValueError(f"Nested properties are not supported for operation {operation}")
            else:
                raise ValueError(f"Unexpected property type: {type(prop)}")
        return output

    def execute_query(
        self,
        view: dm.ViewId,
        operation: Literal["list", "aggregate", "search"],
        properties: Properties | None = None,
        filter: filters.Filter | None = None,
        query: str | None = None,
        group_by: str | SequenceNotStr[str] | None = None,
        aggregates: Aggregation | Sequence[Aggregation] | None = None,
        sort: dm.InstanceSort | Sequence[dm.InstanceSort] | None = None,
        limit: int | None = None,
    ) -> dict[str, Any]:
        warnings.warn(
            "This method is deprecated. Use list, aggregate or search methods instead", UserWarning, stacklevel=2
        )
        dumped: Any
        if operation == "list":
            if properties is None:
                raise ValueError("Properties are required for list operation")
            dumped = self._execute_list(view, properties, filter, sort, limit)
        elif operation == "aggregate":
            if not aggregates:
                raise ValueError("Aggregates are required for aggregate operation")
            flatten_props = self._as_property_list(properties, operation) if properties else None
            dumped = self._execute_aggregation(view, aggregates, flatten_props, query, filter, group_by, limit)
        elif operation == "search":
            flatten_props = self._as_property_list(properties, operation) if properties else None
            search_result = self._client.data_modeling.instances.search(
                view, query, properties=flatten_props, filter=filter, limit=limit or SEARCH_LIMIT, sort=sort
            )
            dumped = self._prepare_list_result(search_result, set(flatten_props) if flatten_props else None)
        else:
            raise NotImplementedError(f"Operation {operation} is not supported")
        return {f"{operation}{view.external_id}": dumped}

    def _execute_list(
        self,
        view_id: dm.ViewId,
        properties: Properties,
        filter: filters.Filter | None = None,
        sort: Sequence[dm.InstanceSort] | dm.InstanceSort | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]:
        view = self._get_view(view_id)
        root_properties = self._as_property_list(properties, "list")
        builder = QueryBuilder()
        factory = QueryStepFactory(view, properties, builder.create_name)

        if not factory.connection_properties:
            result = self._client.data_modeling.instances.list(
                instance_type="node",
                sources=[view_id],
                filter=filter,
                limit=limit,
                sort=sort,
            )
            return self._prepare_list_result(result, set(root_properties))

        reverse_views = {
            prop.through.source: self._get_view(prop.through.source)
            for prop in factory.reverse_properties.values()
            if isinstance(prop.through.source, dm.ViewId)
        }
        builder.append(factory.root(filter, limit))
        for connection_id, connection in factory.connection_properties.items():
            builder.extend(factory.from_connection(connection_id, connection, reverse_views))
        _ = builder.execute_query(self._client, remove_not_connected=False)
        return QueryUnpacker(builder).unpack()

    @classmethod
    def _prepare_list_result(
        cls, result: dm.NodeList[dm.Node], selected_properties: set[str] | None
    ) -> list[dict[str, Any]]:
        output: list[dict[str, Any]] = []
        for node in result:
            item = QueryUnpacker.flatten_dump(node, selected_properties)
            if item:
                output.append(item)
        return output

    def _execute_aggregation(
        self,
        view_id: dm.ViewId,
        aggregates: Aggregation | Sequence[Aggregation],
        search_properties: str | SequenceNotStr[str] | None = None,
        query: str | None = None,
        filter: filters.Filter | None = None,
        group_by: str | SequenceNotStr[str] | None = None,
        limit: int | None = None,
    ) -> dict[str, Any] | list[dict[str, Any]]:
        aggregates_list = aggregates if isinstance(aggregates, Sequence) else [aggregates]
        metric_aggregates = [agg for agg in aggregates_list if isinstance(agg, dm.aggregations.MetricAggregation)]
        histogram_aggregates = [agg for agg in aggregates_list if isinstance(agg, dm.aggregations.Histogram)]
        if metric_aggregates and histogram_aggregates:
            raise ValueError("Cannot mix metric and histogram aggregations")

        if metric_aggregates and group_by is not None:
            group_by_result = self._client.data_modeling.instances.aggregate(  # type: ignore[call-overload]
                view=view_id,
                group_by=group_by,
                aggregates=metric_aggregates,
                query=query,
                properties=search_properties,
                filter=filter,
                limit=limit or AGGREGATION_LIMIT,
            )
            return self._grouped_metric_aggregation_to_dict(group_by_result)
        elif metric_aggregates:
            metric_results = self._client.data_modeling.instances.aggregate(
                view_id,
                aggregates=metric_aggregates,
                query=query,
                properties=search_properties,
                filter=filter,
                limit=limit or AGGREGATION_LIMIT,
            )
            return self._metric_aggregation_to_dict(metric_results)

        elif histogram_aggregates:
            histogram_results = self._client.data_modeling.instances.histogram(
                view_id,
                histograms=histogram_aggregates,
                query=query,
                properties=search_properties,  # type: ignore[arg-type]
                filter=filter,
                limit=limit or AGGREGATION_LIMIT,
            )
            return self._histogram_aggregation_to_dict(histogram_results)
        else:
            raise ValueError("No aggregation found")

    @staticmethod
    def _metric_aggregation_to_dict(aggregation: list[dm.aggregations.AggregatedNumberedValue]) -> dict[str, Any]:
        values_by_aggregations: dict[str, dict[str, Any]] = defaultdict(dict)
        for item in aggregation:
            values_by_aggregations[item._aggregate][item.property] = item.value
        return dict(values_by_aggregations)

    @classmethod
    def _grouped_metric_aggregation_to_dict(cls, aggregations: InstanceAggregationResultList) -> list[dict[str, Any]]:
        output: list[dict[str, Any]] = []
        for group in aggregations:
            group_dict = {
                "group": group.group,
                **cls._metric_aggregation_to_dict(group.aggregates),
            }
            output.append(group_dict)
        return output

    @staticmethod
    def _histogram_aggregation_to_dict(aggregation: list[dm.aggregations.HistogramValue]) -> dict[str, Any]:
        output: dict[str, dict[str, Any]] = defaultdict(dict)
        for item in aggregation:
            output[item._aggregate][item.property] = {
                "interval": item.interval,
                "buckets": [bucket.dump() for bucket in item.buckets],
            }
        return dict(output)

    def search(
        self,
        view: dm.ViewId,
        properties: Properties | None = None,
        query: str | None = None,
        filter: filters.Filter | None = None,
        search_properties: str | SequenceNotStr[str] | None = None,
        sort: Sequence[dm.InstanceSort] | dm.InstanceSort | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]:
        search_result = self._client.data_modeling.instances.search(
            view,
            query,
            properties=search_properties,  # type: ignore[arg-type]
            filter=filter,
            limit=limit or SEARCH_LIMIT,
            sort=sort,
        )

        flatten_props = self._as_property_list(properties, "list") if properties else None
        are_flat_properties = flatten_props == properties
        if properties is None or are_flat_properties:
            return self._prepare_list_result(search_result, set(flatten_props) if flatten_props else None)

        # Lookup nested properties:

        order_by_node_ids = {node.as_id(): no for no, node in enumerate(search_result)}
        # If we are sorting, then we need to ensure externalId and space are included in the properties.
        # This is because we need them for the final sorting.
        include_space = False
        include_external_id = False
        if sort is not None:
            include_space = "space" not in properties
            include_external_id = "externalId" not in properties
        if include_space:
            properties.append("space")
        if include_external_id:
            properties.append("externalId")

        result: list[dict[str, Any]] = []
        for space, space_nodes in itertools.groupby(
            sorted(order_by_node_ids.keys(), key=lambda x: x.space), key=lambda x: x.space
        ):
            is_space = filters.Equals(["node", "space"], space)
            for chunk in chunker(list(space_nodes), IN_FILTER_CHUNK_SIZE):
                batch_filter = filters.And(
                    filters.In(["node", "externalId"], [node.external_id for node in chunk]), is_space
                )
                batch_result = self.list(view, properties, batch_filter, sort, limit or SEARCH_LIMIT)
                result.extend(batch_result)

        if sort is not None:
            result.sort(key=lambda x: order_by_node_ids[dm.NodeId(x["space"], x["externalId"])])
        if include_space or include_external_id:
            for item in result:
                if include_space:
                    del item["space"]
                if include_external_id:
                    del item["externalId"]

        return result

    @overload
    def aggregate(
        self,
        view: dm.ViewId,
        aggregates: Aggregation | Sequence[Aggregation],
        group_by: None = None,
        filter: filters.Filter | None = None,
        query: str | None = None,
        search_properties: str | SequenceNotStr[str] | None = None,
        limit: int | None = None,
    ) -> dict[str, Any]: ...

    @overload
    def aggregate(
        self,
        view: dm.ViewId,
        aggregates: Aggregation | Sequence[Aggregation],
        group_by: str | SequenceNotStr[str],
        filter: filters.Filter | None = None,
        query: str | None = None,
        search_properties: str | SequenceNotStr[str] | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]: ...

    def aggregate(
        self,
        view: dm.ViewId,
        aggregates: Aggregation | Sequence[Aggregation],
        group_by: str | SequenceNotStr[str] | None = None,
        filter: filters.Filter | None = None,
        query: str | None = None,
        search_properties: str | SequenceNotStr[str] | None = None,
        limit: int | None = None,
    ) -> dict[str, Any] | list[dict[str, Any]]:
        return self._execute_aggregation(view, aggregates, search_properties, query, filter, group_by, limit)

    def list(
        self,
        view: dm.ViewId,
        properties: Properties,
        filter: filters.Filter | None = None,
        sort: Sequence[dm.InstanceSort] | dm.InstanceSort | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]:
        return self._execute_list(view, properties, filter, sort, limit)


class QueryStepFactory:
    def __init__(
        self,
        view: dm.View,
        user_selected_properties: Properties,
        create_step_name: Callable[[str | None], str],
    ) -> None:
        self._view = view
        self._view_id = view.as_id()
        self._user_selected_properties = user_selected_properties
        self._create_step_name = create_step_name
        self._root_name: str | None = None

    @cached_property
    def _root_properties(self) -> list[str]:
        root_properties: list[str] = []
        for prop in self._user_selected_properties:
            if isinstance(prop, str):
                root_properties.append(prop)
            elif isinstance(prop, dict):
                key = next(iter(prop.keys()))
                root_properties.append(key)
        return root_properties

    @cached_property
    def connection_properties(self) -> dict[str, ViewProperty]:
        output: dict[str, ViewProperty] = {}
        for prop in self._root_properties:
            definition = self._view.properties.get(prop)
            if not definition:
                continue
            if self._is_connection(definition):
                output[prop] = definition
        return output

    @cached_property
    def reverse_properties(self) -> dict[str, ReverseDirectRelation]:
        return {
            prop_id: prop
            for prop_id, prop in self.connection_properties.items()
            if isinstance(prop, ReverseDirectRelation)
        }

    @cached_property
    def _nested_properties_by_property(self) -> dict[str, list[str | dict[str, list[str]]]]:
        nested_properties_by_property: dict[str, list[str | dict[str, list[str]]]] = {}
        for prop in self._user_selected_properties:
            if isinstance(prop, dict):
                key, value = next(iter(prop.items()))
                nested_properties_by_property[key] = value
        return nested_properties_by_property

    def root(self, filter: dm.Filter | None = None, limit: int | None = None) -> QueryStep:
        skip = _NODE_PROPERTIES | set(self.reverse_properties.keys())
        if self._root_name is not None:
            raise ValueError("Root step is already created")
        self._root_name = self._create_step_name(None)
        has_data = dm.filters.HasData(views=[self._view_id])
        return QueryStep(
            self._root_name,
            dm.query.NodeResultSetExpression(
                filter=dm.filters.And(filter, has_data) if filter else has_data,
            ),
            select=self._create_select([prop for prop in self._root_properties if prop not in skip], self._view_id),
            selected_properties=self._root_properties,
            view_id=self._view_id,
            max_retrieve_limit=-1 if limit is None else limit,
            raw_filter=filter,
        )

    def from_connection(
        self, connection_id: str, connection: ViewProperty, reverse_views: dict[dm.ViewId, dm.View]
    ) -> list[QueryStep]:
        if self._root_name is None:
            raise ValueError("Root step is not created")
        view_property = ViewPropertyId(self._view_id, connection_id)
        selected_properties = self._nested_properties_by_property.get(connection_id, ["*"])
        if isinstance(connection, dm.EdgeConnection):
            return self._from_edge(connection, view_property, selected_properties)
        elif isinstance(connection, ReverseDirectRelation):
            validated = self._validate_flat_properties(selected_properties)
            return self._from_reverse_relation(connection, view_property, validated, reverse_views)
        elif isinstance(connection, dm.MappedProperty) and isinstance(connection.type, dm.DirectRelation):
            validated = self._validate_flat_properties(selected_properties)
            return self.from_direct_relation(connection, view_property, validated)
        else:
            warnings.warn(f"Unexpected connection type: {connection!r}", UserWarning, stacklevel=2)
        return []

    def from_direct_relation(
        self,
        connection: dm.MappedProperty,
        view_property: ViewPropertyId,
        selected_properties: list[str],
    ) -> list[QueryStep]:
        query_properties = self._create_query_properties(selected_properties, None)
        if connection.source is None:
            raise ValueError("Source view not found")
        return [
            QueryStep(
                self._create_step_name(self._root_name),
                dm.query.NodeResultSetExpression(
                    from_=self._root_name,
                    direction="outwards",
                    through=self._view_id.as_property_ref(view_property.property),
                ),
                view_property=view_property,
                select=self._create_select(query_properties, connection.source),
                selected_properties=selected_properties,
                view_id=connection.source,
            )
        ]

    def _from_edge(
        self,
        connection: dm.EdgeConnection,
        view_property: ViewPropertyId,
        selected_properties: list[str | dict[str, list[str]]],
    ) -> list[QueryStep]:
        edge_name = self._create_step_name(self._root_name)
        steps = []
        step = QueryStep(
            edge_name,
            dm.query.EdgeResultSetExpression(
                from_=self._root_name,
                direction=connection.direction,
                chain_to="source" if connection.direction == "outwards" else "destination",
            ),
            selected_properties=[prop for prop in selected_properties if isinstance(prop, str)],
            view_property=view_property,
        )
        steps.append(step)

        node_properties = next(
            (prop for prop in selected_properties if isinstance(prop, dict) and "node" in prop), None
        )
        if isinstance(node_properties, dict):
            selected_node_properties = node_properties["node"]
            query_properties = self._create_query_properties(selected_node_properties, None)
            target_view = connection.source

            step = QueryStep(
                self._create_step_name(edge_name),
                dm.query.NodeResultSetExpression(
                    from_=edge_name,
                    filter=dm.filters.HasData(views=[target_view]),
                ),
                select=self._create_select(query_properties, target_view),
                selected_properties=selected_node_properties,
                view_property=ViewPropertyId(target_view, "node"),
                view_id=target_view,
            )
            steps.append(step)
        return steps

    def _from_reverse_relation(
        self,
        connection: ReverseDirectRelation,
        view_property: ViewPropertyId,
        selected_properties: list[str],
        reverse_views: dict[dm.ViewId, dm.View],
    ) -> list[QueryStep]:
        connection_type: Literal["reverse-list"] | None = (
            "reverse-list" if self._is_listable(connection.through, reverse_views) else None
        )
        query_properties = self._create_query_properties(selected_properties, connection.through.property)
        try:
            other_view = reverse_views[connection.source]
        except KeyError:
            raise ValueError(f"View {connection.source} not found in {reverse_views.keys()}") from None
        return [
            QueryStep(
                self._create_step_name(self._root_name),
                dm.query.NodeResultSetExpression(
                    from_=self._root_name,
                    direction="inwards",
                    through=other_view.as_property_ref(connection.through.property),
                ),
                view_property=view_property,
                select=self._create_select(query_properties, other_view.as_id()),
                selected_properties=selected_properties,
                view_id=connection.source,
                connection_type=connection_type,
            )
        ]

    @classmethod
    def _create_query_properties(cls, properties: list[str], connection_id: str | None = None) -> list[str]:
        include_connection_prop = "*" not in properties
        nested_properties: list[str] = []
        for prop_id in properties:
            if prop_id in _NODE_PROPERTIES:
                continue
            if prop_id == connection_id:
                include_connection_prop = False
            nested_properties.append(prop_id)
        if include_connection_prop and connection_id:
            nested_properties.append(connection_id)
        return nested_properties

    @staticmethod
    def _is_connection(definition: ViewProperty) -> bool:
        return isinstance(definition, dm.ConnectionDefinition) or (
            isinstance(definition, dm.MappedProperty) and isinstance(definition.type, dm.DirectRelation)
        )

    @staticmethod
    def _create_select(properties: list[str], view_id: dm.ViewId) -> dm.query.Select:
        return dm.query.Select([dm.query.SourceSelector(view_id, properties)])

    @staticmethod
    def _is_listable(property: dm.PropertyId, reverse_views: dict[dm.ViewId, dm.View]) -> bool:
        if isinstance(property.source, dm.ViewId):
            try:
                view = reverse_views[property.source]
            except KeyError:
                raise ValueError(f"View {property.source} not found in {reverse_views.keys()}") from None
            if property.property not in view.properties:
                raise TypeError(f"Reverse property {property.property} not found in {property.source!r}")
            reverse_prop = view.properties[property.property]
            if not (isinstance(reverse_prop, dm.MappedProperty) and isinstance(reverse_prop.type, dm.DirectRelation)):
                raise TypeError(f"Property {property.property} is not a direct relation")
            return reverse_prop.type.is_list
        else:
            raise NotImplementedError(f"Property {property.source=} is not supported")

    @staticmethod
    def _validate_flat_properties(properties: list[str | dict[str, list[str]]]) -> list[str]:
        output = []
        for prop in properties:
            if isinstance(prop, str):
                output.append(prop)
            else:
                raise ValueError(f"Direct relations do not support nested properties. Got {prop}")
        return output


class QueryUnpacker:
    def __init__(self, builder: QueryBuilder):
        self._builder = builder

    def unpack(self) -> list[dict[str, Any]]:
        nodes_by_from: dict[str | None, list[tuple[str, dict[dm.NodeId, list[dict[str, Any]]]]]] = defaultdict(list)
        for step in reversed(self._builder):
            source_property = step.view_property and step.view_property.property
            if node_expression := step.node_expression:
                unpacked = self._unpack_node(step, node_expression, nodes_by_from)
                nodes_by_from[step.from_].append((source_property, unpacked))
            elif edge_expression := step.edge_expression:
                step_properties = set(step.selected_properties or [])
                unpacked_edge: dict[dm.NodeId, list[dict[str, Any]]] = defaultdict(list)
                for edge in step.edge_results:
                    start_node = dm.NodeId.load(edge.start_node.dump())
                    end_node = dm.NodeId.load(edge.end_node.dump())
                    dumped = self.flatten_dump(edge, step_properties)
                    if edge_expression.direction == "outwards":
                        source_node = start_node
                        target_node = end_node
                    else:
                        source_node = end_node
                        target_node = start_node
                    for nested_prop, nested_by_id in nodes_by_from.get(step.name, []):
                        if target_node in nested_by_id:
                            dumped[nested_prop] = nested_by_id[target_node]

                    unpacked_edge[source_node].append(dumped)
                nodes_by_from[step.from_].append((source_property, unpacked_edge))
            else:
                raise TypeError("Unexpected step")
        return [item[0] for item in nodes_by_from[None][0][1].values()]

    @classmethod
    def flatten_dump(
        cls, node: dm.Node | dm.Edge, selected_properties: set[str] | None, connection_property: str | None = None
    ) -> dict[str, Any]:
        dumped = node.dump()
        dumped_properties = dumped.pop("properties", {})
        item = {
            key: value for key, value in dumped.items() if selected_properties is None or key in selected_properties
        }
        for _, props_by_view_id in dumped_properties.items():
            for __, props in props_by_view_id.items():
                for key, value in props.items():
                    if key == connection_property:
                        if isinstance(value, dict):
                            item[key] = dm.NodeId.load(value)
                        elif isinstance(value, list):
                            item[key] = [dm.NodeId.load(item) for item in value]
                        else:
                            raise TypeError(f"Unexpected connection property value: {value}")
                    elif selected_properties is None or key in selected_properties:
                        item[key] = value
        return item

    def _unpack_node(
        self,
        step: QueryStep,
        node_expression: dm.query.NodeResultSetExpression,
        results_by_from: dict[str | None, list[tuple[str, dict[dm.NodeId, list[dict[str, Any]]]]]],
    ) -> dict[dm.NodeId, list[dict[str, Any]]]:
        step_properties = set(step.selected_properties or [])
        connection_property: str | None = None
        if node_expression.through and node_expression.direction == "inwards":
            connection_property = node_expression.through.property
        unpacked: dict[dm.NodeId, list[dict[str, Any]]] = defaultdict(list)
        for node in step.node_results:
            node_id = node.as_id()
            dumped = self.flatten_dump(node, step_properties, connection_property)
            for nested_prop, nested_by_id in results_by_from.get(step.name, []):
                if neste_item := nested_by_id.get(node_id):
                    # Reverse or Edge
                    dumped[nested_prop] = neste_item
                elif nested_prop in dumped:
                    # Direct relation
                    identifier = dumped.pop(nested_prop)
                    if isinstance(identifier, dict):
                        other_id = dm.NodeId.load(identifier)
                        if other_id in nested_by_id:
                            dumped[nested_prop] = nested_by_id[other_id]
                    elif isinstance(identifier, list):
                        dumped[nested_prop] = []
                        for item in identifier:
                            other_id = dm.NodeId.load(item)
                            if other_id in nested_by_id:
                                dumped[nested_prop].extend(nested_by_id[other_id])
                else:
                    warnings.warn(
                        f"Nested property {nested_prop} not found in {dumped.keys()}", UserWarning, stacklevel=2
                    )

            if connection_property is None:
                unpacked[node_id].append(dumped)
            else:
                reverse = dumped.pop(connection_property)
                if isinstance(reverse, dm.NodeId):
                    unpacked[reverse].append(dumped)
                elif isinstance(reverse, list):
                    for item in reverse:
                        unpacked[item].append(copy.deepcopy(dumped))
        return unpacked



from collections.abc import Mapping

from cognite.client import data_modeling as dm

_READONLY_PROPERTIES: Mapping[dm.ContainerId, frozenset[str]] = {
    dm.ContainerId("cdf_cdm", "CogniteAsset"): frozenset(
        {"assetHierarchy_root", "assetHierarchy_path", "assetHierarchy_path_last_updated_time"}
    ),
    dm.ContainerId("cdf_cdm", "CogniteFile"): frozenset({"isUploaded", "uploadedTime"}),
}


def is_readonly_property(prop: dm.MappedProperty) -> bool:
    return prop.container in _READONLY_PROPERTIES and prop.name in _READONLY_PROPERTIES[prop.container]



"""This module contains the class representing the filter used in the generated classes."""

from __future__ import annotations

import itertools
from collections.abc import Iterable, Sequence
from dataclasses import dataclass

from cognite.client.data_classes import data_modeling as dm

from cognite.pygen import config as pygen_config
from cognite.pygen.config.reserved_words import is_reserved_word

from .fields import BaseConnectionField, Field, OneToManyConnectionField, OneToOneConnectionField, PrimitiveField


@dataclass
class FilterParameter:
    """This class represents a single parameter in a filter method.

    Args:
        name: The name of the parameter.
        type_: The type of the parameter.
        description: The description of the parameter.
        default: The default value of the parameter.
        is_nullable: Whether the parameter is nullable.
    """

    name: str
    type_: str
    description: str
    default: str | None = None
    is_nullable: bool = True

    def __post_init__(self):
        if is_reserved_word(self.name, "parameter"):
            self.name = f"{self.name}_"

    def copy(self, name: str) -> FilterParameter:
        return FilterParameter(
            name=name,
            type_=self.type_,
            description=self.description,
            default=self.default,
            is_nullable=self.is_nullable,
        )

    @property
    def annotation(self) -> str:
        if self.is_nullable:
            return f"{self.type_} | None"
        else:
            return self.type_

    @property
    def is_time(self) -> bool:
        return self.type_ in ("datetime.datetime", "datetime.date")

    @property
    def is_timestamp(self) -> bool:
        return self.type_ in ("datetime.datetime",)


@dataclass
class FilterImplementation:
    """This class represent a filter implementation.

    It is used in the methods that require filtering to create the DMS filter.

    Args:
        filter: The DMS filter to use.
        prop_name: The name of the property in the View to filter on.
        keyword_arguments: The keyword arguments of the filter.
        is_edge_class: Whether the class is an edge class.
    """

    filter: type[dm.Filter]
    prop_name: str
    keyword_arguments: dict[str, FilterParameter]
    is_edge_class: bool

    @property
    def condition(self) -> str:
        if self.filter is dm.filters.In:
            parameter_name = next(iter(self.keyword_arguments.values())).name
            return f"{parameter_name} and isinstance({parameter_name}, list)"
        elif self.filter is dm.filters.Equals:
            parameter = next(iter(self.keyword_arguments.values()))
            parameter_type = parameter.type_
            if "|" in parameter_type:
                parameter_type = parameter_type.split("|")[0].strip()
            return f"isinstance({parameter.name}, {parameter_type})"

        return " or ".join(f"{arg.name} is not None" for arg in self.keyword_arguments.values())

    @property
    def arguments(self) -> str:
        if self.prop_name in {"externalId", "space"}:
            instance_type = "edge" if self.is_edge_class else "node"
            property_ref = f'["{instance_type}", "{self.prop_name}"], '
        else:
            property_ref = f'view_id.as_property_ref("{self.prop_name}"), '

        filter_args = self._create_filter_args()

        return f"{property_ref}{', '.join(filter_args)}"

    def _create_filter_args(self) -> list[str]:
        filter_args: list[str] = []
        for keyword, arg in self.keyword_arguments.items():
            if arg.is_time:
                timespec = 'timespec="milliseconds"' if arg.is_timestamp else ""
                filter_args.append(f"{keyword}={arg.name}.isoformat({timespec}) if {arg.name} else None")
            else:
                filter_args.append(f"{keyword}={arg.name}")
        return filter_args

    @property
    def filter_call(self) -> str:
        return f"dm.filters.{self.filter.__name__}"


@dataclass
class FilterImplementationOnetoOneEdge(FilterImplementation):
    instance_type: str
    has_default_instance_space: bool

    @property
    def condition(self) -> str:
        if self.filter is dm.filters.In:
            parameter = next(iter(self.keyword_arguments.values())).name
            return (
                f"{parameter} and isinstance({parameter}, Sequence) "
                f"and not isinstance({parameter}, str) and not is_tuple_id({parameter})"
            )
        elif self.filter is dm.filters.Equals:
            parameter = next(iter(self.keyword_arguments.values())).name
            return f"isinstance({parameter}, {self.instance_type}) or is_tuple_id({parameter})"
        raise NotImplementedError(f"Unsupported filter {self.filter} for Direct Relation")

    def _create_filter_args(self) -> list[str]:
        filter_args: list[str] = []
        for keyword, arg in self.keyword_arguments.items():
            if self.instance_type == Sequence.__name__ and self.filter is dm.filters.In:
                filter_args.append(f"{keyword}=[as_instance_dict_id(item) for item in {arg.name}]")
            else:
                filter_args.append(f"{keyword}=as_instance_dict_id({arg.name})")
        return filter_args


@dataclass
class FilterMethod:
    """This classe represents a filter that can be used in any method that requires filtering.

    Args:
        parameters: The parameters of the filter.
        implementations: The conditions of the filter.

    """

    parameters: list[FilterParameter]
    implementations: list[FilterImplementation]

    @classmethod
    def from_fields(
        cls,
        fields: Iterable[Field],
        config: pygen_config.Filtering,
        has_default_instance_space: bool,
        is_edge_class: bool = False,
    ) -> FilterMethod:
        parameters_by_name: dict[str, FilterParameter] = {}
        list_filters: list[FilterImplementation] = []

        for field_ in itertools.chain(fields, (_EXTERNAL_ID_FIELD, _SPACE_FIELD)):
            if isinstance(field_, BaseConnectionField) and field_.is_reverse_direct_relation:
                # Reverse direct relations are not supported in list methods.
                continue
            # Only primitive and edge one-to-one fields supported for now
            if isinstance(field_, PrimitiveField):
                for selected_filter in config.get(field_.type_, field_.prop_name):
                    if selected_filter is dm.filters.Equals:
                        if field_.name not in parameters_by_name:
                            parameter = FilterParameter(
                                name=field_.name,
                                type_=field_.type_as_string,
                                description=f"The {field_.doc_name} to filter on.",
                            )
                            parameters_by_name[parameter.name] = parameter
                        else:
                            # Equals and In filter share parameter, you have to extend the type hint.
                            parameter = parameters_by_name[field_.name]
                            parameter.type_ = f"{field_.type_as_string} | {parameter.type_}"
                        list_filters.append(
                            FilterImplementation(
                                filter=selected_filter,
                                prop_name=field_.prop_name,
                                keyword_arguments=dict(value=parameter),
                                is_edge_class=is_edge_class,
                            )
                        )
                    elif selected_filter is dm.filters.In:
                        if field_.name not in parameters_by_name:
                            parameter = FilterParameter(
                                field_.name,
                                type_=f"list[{field_.type_as_string}]",
                                description=f"The {field_.doc_name} to filter on.",
                            )
                            parameters_by_name[parameter.name] = parameter
                        else:
                            # Equals and In filter share parameter, you have to extend the type hint.
                            parameter = parameters_by_name[field_.name]
                            parameter.type_ = f"{parameter.type_} | list[{field_.type_as_string}]"
                        list_filters.append(
                            FilterImplementation(
                                filter=selected_filter,
                                prop_name=field_.prop_name,
                                keyword_arguments=dict(values=parameter),
                                is_edge_class=is_edge_class,
                            )
                        )
                    elif selected_filter is dm.filters.Prefix:
                        parameter = FilterParameter(
                            name=f"{field_.name}_prefix" if field_.name[-1] != "_" else f"{field_.name}prefix",
                            type_=field_.type_as_string,
                            description=f"The prefix of the {field_.doc_name} to filter on.",
                        )
                        parameters_by_name[parameter.name] = parameter
                        list_filters.append(
                            FilterImplementation(
                                filter=selected_filter,
                                prop_name=field_.prop_name,
                                keyword_arguments=dict(value=parameter),
                                is_edge_class=is_edge_class,
                            )
                        )
                    elif selected_filter is dm.filters.Range:
                        min_parameter = FilterParameter(
                            name=f"min_{field_.name}",
                            type_=field_.type_as_string,
                            description=f"The minimum value of the {field_.doc_name} to filter on.",
                        )
                        max_parameter = FilterParameter(
                            name=f"max_{field_.name}",
                            type_=field_.type_as_string,
                            description=f"The maximum value of the {field_.doc_name} to filter on.",
                        )
                        parameters_by_name[min_parameter.name] = min_parameter
                        parameters_by_name[max_parameter.name] = max_parameter
                        list_filters.append(
                            FilterImplementation(
                                filter=selected_filter,
                                prop_name=field_.prop_name,
                                keyword_arguments=dict(gte=min_parameter, lte=max_parameter),
                                is_edge_class=is_edge_class,
                            )
                        )
                    else:
                        # This is a filter not supported by the list method.
                        continue
            elif isinstance(field_, OneToOneConnectionField | OneToManyConnectionField) and field_.is_direct_relation:
                for selected_filter in config.get(dm.DirectRelation(), field_.prop_name):
                    if selected_filter is dm.filters.Equals:
                        type_ = "tuple[str, str] | dm.NodeId | dm.DirectRelationReference"
                        if has_default_instance_space:
                            type_ = f"str | {type_}"
                        if field_.name not in parameters_by_name:
                            parameter = FilterParameter(
                                name=field_.name,
                                type_=type_,
                                description=f"The {field_.doc_name} to filter on.",
                            )
                            parameters_by_name[parameter.name] = parameter
                        else:
                            # Equals and In filter share parameter, you have to extend the type hint.
                            parameter = parameters_by_name[field_.name]
                            parameter.type_ = f"{type_} | {parameter.type_}"
                        instance_type = "dm.NodeId | dm.DirectRelationReference"
                        if has_default_instance_space:
                            instance_type = f"str | {instance_type}"

                        list_filters.append(
                            FilterImplementationOnetoOneEdge(
                                filter=selected_filter,
                                prop_name=field_.prop_name,
                                keyword_arguments=dict(value=parameter),
                                instance_type=instance_type,
                                is_edge_class=is_edge_class,
                                has_default_instance_space=has_default_instance_space,
                            )
                        )
                    elif selected_filter is dm.filters.In:
                        type_ = "tuple[str, str] | dm.NodeId | dm.DirectRelationReference"
                        if has_default_instance_space:
                            type_ = f"str | {type_}"
                        type_ = f"Sequence[{type_}]"

                        if field_.name not in parameters_by_name:
                            parameter = FilterParameter(
                                name=field_.name,
                                type_=type_,
                                description=f"The {field_.doc_name} to filter on.",
                            )
                            parameters_by_name[parameter.name] = parameter
                        else:
                            # Equals and In filter share parameter, you have to extend the type hint.
                            parameter = parameters_by_name[field_.name]
                            parameter.type_ = f"{parameter.type_} | {type_}"
                        list_filters.append(
                            FilterImplementationOnetoOneEdge(
                                filter=selected_filter,
                                prop_name=field_.prop_name,
                                keyword_arguments=dict(values=parameter),
                                instance_type=Sequence.__name__,
                                is_edge_class=is_edge_class,
                                has_default_instance_space=has_default_instance_space,
                            )
                        )
                    else:
                        # This is a filter not supported.
                        continue

        return cls(parameters=list(parameters_by_name.values()), implementations=list_filters)


# These fields are used when creating the list method.
_EXTERNAL_ID_FIELD = PrimitiveField(
    name="external_id",
    prop_name="externalId",
    type_=dm.Text(),
    doc_name="external ID",
    is_nullable=False,
    default=None,
    description=None,
    pydantic_field="Field",
)
_SPACE_FIELD = PrimitiveField(
    name="space",
    prop_name="space",
    type_=dm.Text(),
    doc_name="space",
    is_nullable=False,
    default=None,
    description=None,
    pydantic_field="Field",
)
